/* 
	Simulator with cache for Pep/8 as defined in 
        "Computer Systems" by J. Stanley Warford 
	(3rd Edition) Jones and Bartlett 2005.

	Copyright (C) 2006 Peter D. Smith

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation (http://www.gnu.org); either version 2 of 
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    Author:     peter.smith@csuci.edu

		Peter Smith
		Computer Science
		CSU Channel Islands
		One University Drive
		Camarillo, CA 93012
		USA


4/12/05	Version Beta 3.0   4/12/05
        Breakpoints, single-step

9/23/05: OS file without leading zeros is pep8os.pepo
       : Version 3.1 - separate loados function
       : clarified user prompts

11/23 : Version 3.2 - restrict clear and memory modification to Application Program area
		      add command 'a' to reset SP and PC

12/5/05: Version 3.3 - 3.1 with improved user interface by Stan Warford

12/8/05: Version 3.4 - version 3.3 with version 3.2 fixes.

6/8/06:  Version 3.5 - added profiling

6/14/06: Version 3.5.1 - improved command input (strings). Minor changes to interface.

6/23/06: Renamed pep8cache (prompt changed). Changes by Stan Warford.                       
         
6/23/06: Version 3.5.2 (temporarily) removed option to have no cache
                        user must select a replacement algorithm
                        added GNU GPL language
                        modified cache stats output to show total reads (6/27/06)
*/

#include <stdio.h>

#define VERSION "3.5.2"
#define MEMSIZE 65536
#define NUMINSTR 39
#define HALTOPCODE 0
#define TRUE 1
#define FALSE 0
#define CBIT 1
#define VBIT 2
#define ZBIT 4
#define NBIT 8
#define USER 0
#define SYSTEM 1
#define RANDOM 0
#define FIFO 1
#define LFU 2
#define LRU 3
#define MAXBREAKPOINTS 10
#define TRACEBIT1  1
#define TRACEBIT2  2
#define TRACEBIT4  4
#define TRACEBIT8  8
#define TRACEBIT16 16
#define TRACEBIT32 32

// for profile list

typedef struct proflistnode PROFLISTNODE;

struct proflistnode
{
 int linenumber;
 int count;
 PROFLISTNODE *next;
};


unsigned char mem[MEMSIZE];

short RegA, RegX;             // -32768 .. 32767
unsigned short  SP, PC;       // 0..65535
unsigned int WRITELIMIT;      // set by loados - address of first byte of OS
unsigned char Status;         // 0000NZVC
unsigned char Sysmode;        // will be USER or SYSTEM

int iscache;                  // indicates if user has created memory cache

int clocktime;                // used to tag cache block usage

/* cache parameters and structures */

int blockbits, bpr, totalblocks, numrows, rowbits, blocksize, sepid;
int algorithm;   // block replacement algorithm

unsigned char *inuse;
unsigned long *usage;  
unsigned int  *tags;
unsigned char *cache;

unsigned long ireadcount, ireadhits, dreadcount, dreadhits, wcount, whits;

int instructioncounts[NUMINSTR];   // frequency counters

int decodetable [] = {
				 0, 1, 2, 3, 4, 4, 5, 5, 6, 6,
                                 7, 7, 8, 8, 9, 9,10,10,11,11,
                                12,12,13,13,14,14,15,15,16,16,
                                17,17,18,18,19,19,20,20,20,20,
                                21,-1,-1,-1,-1,-1,-1,-1,-1,22,
                                22,22,22,22,22,22,23,23,23,23,
                                23,23,23,23,-1,24,24,-1,24,-1,
                                -1,-1,-1,25,25,25,25,25,25,25,
                                26,26,26,26,26,26,26,26,27,27,
                                27,27,27,27,27,27,28,28,28,28,
                                28,28,28,28,29,29,29,29,29,29,
                                29,29,30,30,30,30,30,30,30,30,
                                30,30,30,30,30,30,30,30,31,31,
                                31,31,31,31,31,31,31,31,31,31,
                                31,31,31,31,32,32,32,32,32,32,
                                32,32,32,32,32,32,32,32,32,32,
                                33,33,33,33,33,33,33,33,33,33,
                                33,33,33,33,33,33,34,34,34,34,
                                34,34,34,34,34,34,34,34,34,34,
                                34,34,35,35,35,35,35,35,35,35,
                                35,35,35,35,35,35,35,35,36,36,
                                36,36,36,36,36,36,36,36,36,36,
                                36,36,36,36,-1,37,37,37,37,37,
                                37,37,-1,37,37,37,37,37,37,37,
                                -1,38,38,38,38,38,38,38,-1,38,
                                38,38,38,38,38,38
			};

int memtrace, instructiontrace, pctrace, tracenumber, ostrap, singlestep, numbreakpoints;
unsigned short SPHWM;   // stack pointer high water mark
FILE *memtracefile, *instructiontracefile;
unsigned int breakpoints[MAXBREAKPOINTS];
PROFLISTNODE *L;




/* prototypes */

void display(unsigned short,unsigned short);         /* display memory */
void clearmem();                                     /* clears the Application Program area of the memory */
int loados();                                        /* load operating system */
int load(char[]);                                    /* load object program from named file */
int convert(char,char);                              /* check and convert hex pair */
int figurerownumber(int,char,int,int);               /* find cache row given address */
unsigned char getfromcache(int,char);                /* retrieve item from cache */
void writetecache(int,char,unsigned char);            /* write operation to byte in cache */
void putincache(int,char);                           /* put new block in cache */
int isincache(int,char);                             /* check if myte is in cache */
unsigned char readfrommem(char,unsigned short);      /* user program read mechanism */
void writetomem(char,unsigned short,unsigned char);  /* user program write mechanism */
void printregisters();                               /* print registers */
void newrun();                                       /* "run" w/single-step, breakpoints */
int run();                                           /* run the program */
int decode(int,int*,int*,int*,int*,int*,int*);       /* decode instruction */
int temptrace(int,int,int,int,int,int,int);          /* diagnostic printing */
int getoperand1(int,unsigned short);                 /* get 1-byte operand */
int getoperand2(int,unsigned short);                 /* get 2-byte operand */
unsigned short effectiveaddress(int,short);          /* for store instructions */
unsigned char msb(unsigned int);                     /* most significant byte */
unsigned char lsb(unsigned int);                     /* least significant byte */
int execute(int,int,int,int,int,int,int);            /* execute instruction */
void printcachestats();                              /* display cache stats */
void displayfreqs();                                 /* frequencies of instructions */
void zeroout();                                      /* zeros the frequency counts */
int isinbreakpoints(unsigned int);                   /* checks breakpoint list for address */
void breakpointhandler();                            /* submenu for breakpoint operations */ 
void initializecache();                              /* initialize (clear) cache */
void hardwaresetup();                                /* optionally sets up cache */
void menu();                                         /* display menu of options */



unsigned char msb (unsigned int W)
{
 return (unsigned char)( (W >> 8) & 0xff);
}


unsigned char lsb (unsigned int W)
{
 return (unsigned char)( W  & 0xff);
}



void display(unsigned short start, unsigned short finish)
{
 /* display memory in rows of 16 hex bytes from row containing start
    to row containing finish
 */
	unsigned short s=(start/16)*16;
	unsigned short f=(finish/16)*16;
	unsigned short row, col;

	printf("\n Address   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F\n");
	if (s<=f)
		{
                 row = s;
                 while (row <=f)
                      {
			printf("     %.3x", row);
			for(col=0; col<=15; col++) printf("  %.2x",mem[row+col]);
			printf("\n");
                        row = row+16;
                        if (row==0) break; // wrapped
                      }
		}
        else
        	printf("\n Invalid Input");
}



void clearmem()
{
 // set all bytes in Application Program area to zero
 int ptr;

 for (ptr=0; ptr<WRITELIMIT; ptr++) mem[ptr]=0;
}



int convert(char msd, char lsd)
{
 /* return value of 2-char hex number or -1 if invalid */

 int msdval, lsdval;

 if (msd >= '0' && msd <= '9') msdval = msd-'0';
 else if (msd >='A' && msd <='F') msdval = msd-'A' + 10;
 else if (msd >='a' && msd <='f') msdval = msd-'a' + 10;
    else return -1;

 if (lsd >= '0' && lsd <= '9') lsdval = lsd-'0';
 else if (lsd >='A' && lsd <='F') lsdval = lsd-'A' + 10;
 else if (lsd >='a' && lsd <='f') lsdval = lsd-'a' + 10;
    else return -1;
 
 return msdval * 16 + lsdval;
}


int loados()
{
 /* load operating system from "pep8os.pepo"
    
    Pass 1 - check file and count bytes
           - set WRITELIMIT according to size of OS
    Pass 2 - load OS -last byte at FFFF
 */
    char first,second,third;
    int moredata,memorypointer,temp;
    unsigned bytecount;

    FILE *fp;

    fp = fopen("pep8os.pepo","r");
    if (fp==NULL)
      { 
       printf("unable to open pep8os.pepo\n");
       return FALSE;
      }
    else
      {
        /* file opened OK. */

        /* Pass 1 check and count hex pairs separated by space or newline ending in zz */

        bytecount = 0; moredata = TRUE;
        while (moredata)
         {
          fscanf(fp,"%c%c%c",&first,&second,&third);
          if (first =='z' && second == 'z')
               moredata = FALSE ;
          else
             { 
               temp = convert(first,second);
               if (temp != -1)
                 {
                     bytecount++;
                     if (third=='\x0D') fscanf(fp,"%c",&third); /* consume LF after CR */
                 }
                else 
                 {
                      printf("bad data %c%c %d\n",first,second,memorypointer); 
                      return FALSE;  /* failed to load OS */
                 }
              }
          }
 
        /* set WRITELIMIT and memorypointer according to size of OS */

        WRITELIMIT = 65535-bytecount+1;
        memorypointer = WRITELIMIT; 

        fclose(fp);
        fp = fopen("pep8os.pepo","r");

        /* Pass 2 - load OS */

        moredata = TRUE;
        while (moredata)
         {
          fscanf(fp,"%c%c%c",&first,&second,&third);
          if (first =='z' && second == 'z')
               moredata = FALSE ;
          else
             { 
               temp = convert(first,second);
               mem[memorypointer++]=temp;
               if (third=='\x0D') fscanf(fp,"%c",&third); /* consume LF after CR */
             }
         }
     
       fclose(fp);
       /* check load ends ok */

       if (memorypointer & 0xffff !=0)
         return FALSE; 
       else      
       return TRUE; /* successful load of OS */
      }
}


int load(char filename[])
{
    /* loads program from file. If OK returns count of bytes loaded  */
    /* need to add check for missing terminator */

    char first,second,third;
    int moredata,memorypointer,temp;

    FILE *fp;

    fp = fopen(filename,"r");
    if (fp==NULL)
      { 
       printf("load: unable to open %s\n",filename);
       return 0; // zero bytes loaded
      }
    else
      {
        /* file opened OK.  load hex pairs separated by space or newline ending in zz */

        memorypointer = 0; moredata = TRUE;
        while (moredata)
         {
          fscanf(fp,"%c%c%c",&first,&second,&third);
          if (first =='z' && second == 'z')
               moredata = FALSE ;
          else
             { 
               temp = convert(first,second);
               if (temp != -1)
                 {
                     mem[memorypointer]=temp;
                     /*printf("just assigned %d to mem %d\n",temp,memorypointer);*/
                     memorypointer++;
                     if (third=='\x0D') fscanf(fp,"%c",&third); /* consume LF after CR - 9/8/05 */
                 }
                else 
                 {
                      printf("load: bad data %c%c at byte %d\n",first,second,memorypointer); 
                      return 0;  /* return zero bytes loaded */
                 }
              }
          }

       // following register initializations are so that "run" works as single-step operation
        PC = 0;  RegA = 0; RegX = 0;  Status = 0;
        SP = (mem[0xFFF8]<<8)|mem[0xFFF9];  // OS pointer to base of user stack.
        SPHWM = SP;

        return memorypointer; /* count of bytes */
      }
}



int figurerownumber (int address, char iord, int rowbits, int blockbits)
{
  // figure out what row in the cache the address maps to. Depends on
  // how many bits of address speicify row
  // how many bits of address specifiy byte-within-block
  // whether separate i and d caches

 int result;

  if (rowbits==0)
         result = 0;
  else
       {          
          if (!sepid)
        {
          // i and d mixed in cache     
          result =  (address>>blockbits) & (~(0xffff << rowbits) ) ;
        }
     else
        {
          // separate rows for i and for d
           if (rowbits==1)   // only 1 possible row for each data type
                result = 0;
           else result = ( (address>>blockbits) & (~(0xffff << (rowbits-1)) ) ) ;
          // i is first half of rows; d is second half of rows
          if (iord=='d') result += (1 << (rowbits-1));
        }
     }

 if ((tracenumber & TRACEBIT16) !=0)
   printf("figurerow: address:%x iord:%c rowbits:%d  blockbits:%d maps to row%d\n",address,iord,rowbits,blockbits,result);
return result;
}



void putincache(int address, char iord)
{
 // need to figure which block to overwrite then
 // bring in block from memory
 // no copy back at this stage

 int i,j,tagfromaddress,rownumber,target,temp,blockbase;

 if (!sepid) 
             tagfromaddress = address >> (blockbits+rowbits);
      else   tagfromaddress = address >> (blockbits+rowbits-1);

 rownumber = figurerownumber(address,iord,rowbits,blockbits);

 for (i= rownumber*bpr; i< (rownumber+1)*bpr; i++) 
   if (!inuse[i])
       {
         tags[i]=tagfromaddress;
         if (algorithm==FIFO) usage[i]=clocktime;
         inuse[i]=TRUE;
         blockbase = address/blocksize*blocksize;
         if ((tracenumber & TRACEBIT32) != 0)
            printf("putincache: loading block of size %d starting at byte %d\n",blocksize,blockbase);
         for (j=0; j<blocksize; j++) cache[i*blocksize+j ] = mem [blockbase+j];  // memory block brought in  
         return;
       }

 // *** replacement goes here
 
 if (algorithm==RANDOM)
       {
        target = (rownumber*bpr) + (rand()%bpr);
        tags[target]=tagfromaddress;
       }
 if (algorithm==FIFO)
       {
        // look for lowest number in "usage" - the load time
        target = rownumber*bpr;
        temp = usage[target];
        for (i=target+1; i<(rownumber+1)*bpr; i++)
            if (usage[i]<temp) { temp=usage[i]; target=i; }
        tags[target] = tagfromaddress;
        usage[target] = clocktime;
       }
 if (algorithm==LFU)
       {
        // look for lowest number in "usage" - the number of accesses
        target = rownumber*bpr;
        temp = usage[target];
        for (i=target+1; i<(rownumber+1)*bpr; i++)
            if (usage[i]<temp) { temp=usage[i]; target=i; }
        tags[target] = tagfromaddress;
        usage[target] = 0;  // initial usage count
       }
 if (algorithm==LRU)
       {
        // look for lowest number in "usage" - the oldest access
        target = rownumber*bpr;
        temp = usage[target];
        for (i=target+1; i<(rownumber+1)*bpr; i++)
            if (usage[i]<temp) { temp=usage[i]; target=i; }
        tags[target] = tagfromaddress;
        usage[target] = clocktime;  // initial usage time
       }
 // now load block
         blockbase = address/blocksize*blocksize;
         if ((tracenumber & TRACEBIT32) !=0)
            printf("putincache: loading block of size %d starting at byte %d\n",blocksize,blockbase);
         for (j=0; j<blocksize; j++) 
         cache[target*blocksize+j ] = mem [blockbase+j];  // memory block brought in  

  if ((tracenumber & TRACEBIT32) != 0)
   printf("algorithm %d replaced %d at %d in row %d\n",algorithm,target,clocktime,rownumber);
}




unsigned char getfromcache(int address, char iord) 
{
 // item known to be in cache just have to return appropriate unsigned char.
 int i,j,tagfromaddress,rownumber;

 if (!sepid) 
             tagfromaddress = address >> (blockbits+rowbits);
      else   tagfromaddress = address >> (blockbits+rowbits-1);

 rownumber = figurerownumber(address,iord,rowbits,blockbits);

 for (i= rownumber*bpr; i< (rownumber+1)*bpr; i++) 
   if (tags[i]==tagfromaddress && inuse[i]) 
        {
          if ((tracenumber & TRACEBIT16) !=0)
           printf("getfromcache: cache[%d] is %d\n", i * blocksize + ( address & ~(0xffff << blockbits) ),
                   cache [ i * blocksize + ( address & ~(0xffff << blockbits) ) ]);
             return  cache [ i * blocksize + ( address & ~(0xffff << blockbits) ) ];
        }
}



int isincache(int address,char iord)
{
 int i,j,tagfromaddress,rownumber;

 if (!sepid) 
             tagfromaddress = address >> (blockbits+rowbits);
      else   tagfromaddress = address >> (blockbits+rowbits-1);

 rownumber = figurerownumber(address,iord,rowbits,blockbits);

 if (( tracenumber & TRACEBIT16) !=0)
   { 
    printf("isincache: address=%4x rownumber=%x tag=%x " ,address,rownumber,tagfromaddress);
    printf("blockbits = %d rowbits = %d\n",blockbits,rowbits);
   }

 for (i= rownumber*bpr; i< (rownumber+1)*bpr; i++) 
   if (tags[i]==tagfromaddress && inuse[i]) 
        {
          // update usage here
          if (algorithm==LFU)
              {
               usage[i]++;
              }
          if (algorithm==LRU)
              {
               usage[i]=clocktime;
              }
          // printf("isincache: returning true\n");
          return TRUE;
        }

 if ((tracenumber & TRACEBIT16) !=0) printf("isincache: returning false\n");
 return FALSE; 
}




unsigned char readfrommem(char memtype,unsigned short address)
/* user program read mechanism */
{
  if (memtrace) fprintf(memtracefile,"%cR%.4x\n",memtype,address);   

  if (memtype=='i') ireadcount++;
  if (memtype=='d') dreadcount++;

  // see if cache

  if (!iscache)  
    {
        return mem[address];  // no cache just get from memory
    }
  else
    {
     // cache action here
     clocktime++;             // may revise location of this later
     if (isincache(address,memtype))
        {
         if (memtype=='i') ireadhits++;
         if (memtype=='d') dreadhits++;
         return getfromcache(address,memtype);
        }
     else
        {
         putincache(address,memtype);
         return mem[address];
        }
    }
}


void writetocache(int address, char iord, unsigned char data) 
{
 // item known to be in cache just have to write to appropriate location.
 int i,j,tagfromaddress,rownumber;

 if (!sepid) 
             tagfromaddress = address >> (blockbits+rowbits);
      else   tagfromaddress = address >> (blockbits+rowbits-1);

 rownumber = figurerownumber(address,iord,rowbits,blockbits);

 for (i= rownumber*bpr; i< (rownumber+1)*bpr; i++) 
   if (tags[i]==tagfromaddress && inuse[i]) 
        {
         if ((tracenumber & TRACEBIT16) != 0)
             printf("writetocache: cache[%d] data=%d\n", i * blocksize + ( address & ~(0xffff << blockbits) ),data);
         cache [ i * blocksize + ( address & ~(0xffff << blockbits) ) ] = data;
        }
}





void writetomem(char memtype, unsigned short  address, unsigned char data)
/* user program write mechanism */
{
  if ( (Sysmode==SYSTEM) || (address < WRITELIMIT))
     {
       if (memtrace) fprintf(memtracefile,"%cW%.4x\n",memtype,address);   

       wcount++;
       // see if cache       
       if (!iscache)
          {
            mem[address] = data;  // no cache - write to memory
          }
       else
          {
            // do cache stuff here
               clocktime++;
               if (isincache(address,memtype))
                  {
                    whits++;
                    writetocache(address,memtype,data);
                  }    
            // write through strategy
            mem[address] = data; 
          }
     }
  else printf("write to address %d (%x) blocked\n",address,address);
}


int decode(int IR,int *opcode,int *istrap,int *unary,int *reg, int *mode,int  *other)
{
 switch (decodetable[IR])
     {
        case -1:  printf("decode: invalid opcode %d\n",IR);
                  return FALSE;

        case 0:
        case 1:
        case 2:
        case 3:   *opcode = IR; *istrap = 0; *unary = 1;
                  *reg = -1; *mode = -1; *other = -1;
                  break;

        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
        case 9:
        case 10:
        case 11:
        case 12:
        case 13:  *opcode = IR >> 1; *istrap = 0; *unary = 0; *reg = -1;
                  *mode = IR&1; *other = -1;
                  break;

        case 14: 
        case 15: 
        case 16: 
        case 17: 
        case 18: 
        case 19:  *opcode = IR >>1; *istrap = 0; *unary = 1;
                  *reg = IR & 1; *mode = -1; *other = -1; 
                  break;

        case 20:  *opcode = IR >> 2; *istrap = 1; *unary = 1;
                  *reg = -1; *mode = -1; *other = IR & 3; 
                  break;

        case 21:  
        case 22:  
        case 23:  
        case 24:  *opcode = IR >> 3; *istrap = 1; *unary = 0;
                  *reg = -1; *mode = IR&7; *other = -1; 
                  break;

        case 25:  
        case 26:
        case 28:
        case 29:  *opcode = IR >> 3; *istrap = 0; *unary = 0;
                  *reg = -1; *mode = IR&7; *other = -1; 
                  break;

        case 27:  *opcode = IR >>3; *istrap = 0; *unary = 0;
                  *reg = -1; *mode = -1; *other = IR&7; 
                  break;

        case 30:
        case 31:
        case 32:
        case 33:
        case 34:
        case 35:
        case 36:
        case 37:
        case 38:  *opcode = IR >>4; *istrap = 0; *unary = 0;
                  *reg = (IR >> 3) & 1 ; *mode = IR&7; *other = -1;
                  break;
     }
 return TRUE;
}



int temptrace(int opcode, int istrap, int unary, int reg, int mode, int other, int operand)
{
 printf("opcode:%d ", opcode);
 if (reg == -1) printf("      ");  else printf ("reg: %d ",reg);
 if (mode == -1) printf("       ");  else printf ("mode: %d ",mode);
 if (other == -1) printf("       ");  else printf ("other: %d ",other);
 if (!unary) printf("operand: %d",operand);
printf("\n");
return 1;
}




int getoperand1 (int mode, unsigned short operand)
{
 int destination;

     switch (mode)
       { 
          case 0: destination = operand; break;                                                              // i
          case 1: destination = readfrommem('d',operand); break;                                             // d
          case 2: destination = readfrommem('d',operand)<<8 | readfrommem('d',operand+1);                    // n
                  destination = readfrommem('d',destination); break;
          case 3: destination = readfrommem('d',SP+operand); break;                                          // s
          case 4: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1);              // sf
                  destination = readfrommem('d',destination); break;
          case 5: destination = readfrommem('d',RegX+operand); break;                                        // x
          case 6: destination = readfrommem('d',SP+RegX+operand); break;                                     // sx
          case 7: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1);              // sxf
                  destination = readfrommem('d',destination+RegX); break;
       } 
 return destination;
}


int getoperand2 (int mode, unsigned short operand)
{
 int destination;

     switch (mode)
       { 
          case 0: destination = operand; break;                                                                  // i
          case 1: destination = readfrommem('d',operand)<<8 | readfrommem('d',operand+1); break;                 // d
          case 2: destination = readfrommem('d',operand)<<8 | readfrommem('d',operand+1);                        // n
                  destination = readfrommem('d',destination)<<8 | readfrommem('d',destination+1); break;
          case 3: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1); break;           // s
          case 4: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1);                  // sf
                  destination = readfrommem('d',destination)<<8 | readfrommem('d',destination+1); break;
          case 5: destination = readfrommem('d',RegX+operand)<<8 | readfrommem('d',RegX+operand+1); break;       // x
          case 6: destination = readfrommem('d',SP+RegX+operand)<<8 | readfrommem('d',SP+RegX+operand+1); break; // sx
          case 7: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1);                  // sxf
                  destination = readfrommem('d',destination+RegX)<<8 | readfrommem('d',destination+RegX+1); break;
       }
 return destination;
}


unsigned short effectiveaddress(int mode, short operand)
{
 unsigned short destination;

     switch (mode)
       { 
          case 0: printf("effectiveaddress: error mode 0\n");  break;                                            // i
          case 1: destination = operand; break;                                                                  // d
          case 2: destination = readfrommem('d',operand)<<8 | readfrommem('d',operand+1); break;                 // n
          case 3: destination = SP+operand; break;                                                               // s
          case 4: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1); break;           // sf
          case 5: destination = RegX+operand; break;                                                             // x
          case 6: destination = SP+RegX+operand; break;                                                          // sx
          case 7: destination = readfrommem('d',SP+operand)<<8 | readfrommem('d',SP+operand+1);                  // sxf
                  destination = RegX + destination;
       }

 return destination;
}




int execute(int IR, int istrap, int unary, int reg, int mode, int other, int operand)
{
  // prelim version 2/11/05
  // return TRUE if program to continue, FALSE if program to stop.
 short destination;
 char buffer[50];
 char inchar;
 unsigned short memdestination;
 unsigned char tempchar;
 short tempreg;
 int tempa, tempb, tempc;   // temporary 32-bit register for checking C and V conditions

 instructioncounts[decodetable[IR]]++;

 switch (decodetable[IR])
     {
        case -1:  printf("execute: invalid opcode %d\n",IR);
                  return FALSE;

        case 0:   return FALSE;                   // STOP
        
        case 1:  // return from trap              // RETTR  - see p. 304 
                 Status = readfrommem('d',SP);
                 RegA = readfrommem('d',SP+1) << 8 | readfrommem('d',SP+2);
                 RegX = readfrommem('d',SP+3) << 8 | readfrommem('d',SP+4);
                 PC   = readfrommem('d',SP+5) << 8 | readfrommem('d',SP+6);
                 SP   = readfrommem('d',SP+7) << 8 | readfrommem('d',SP+8);
                 Sysmode = USER ;       // return to user mode                 
                 return TRUE;

        case 2:  RegA = SP;                       // MOVSPA
                 return TRUE;

        case 3:  RegA = Status;                   // MOVFLGA
                 return TRUE;

        case 4: if (mode==0)
			destination = operand;               // d
                else    destination = getoperand2(5,operand);// x
                PC = destination;                            //  BR
                return TRUE;

        case 5: if (mode==0)
			destination = operand;               // d
                else    destination = getoperand2(5,operand);// x
                if (Status&NBIT || Status&ZBIT)  PC = destination;  // BRLE
                return TRUE;

        case 6: if (mode==0)
			destination = operand;               // d
                else    destination = getoperand2(5,operand);// x
                if (Status&NBIT)  PC = destination;          //  BRLT
                return TRUE;

        case 7: if (mode==0)
			destination = operand;               // d
                else    destination = getoperand2(5,operand);// x
                if (Status&ZBIT)  PC = destination;          //  BREQ
                return TRUE;

        case 8: if (mode==0)
			destination = operand;               // d
                else    destination = getoperand2(5,operand);// x
                if (!(Status&ZBIT))  PC = destination;       //  BRNE
                return TRUE;

        case 9: if (mode==0)
			destination = operand;               // d
                else    destination = getoperand2(5,operand);// x
                if (!(Status&NBIT))  PC = destination;       //  BRGE
                return TRUE;

        case 10: if (mode==0)
			destination = operand;                // d
                 else    destination = getoperand2(5,operand);// x
                 if (!(Status&ZBIT) && !(Status&NBIT))  PC = destination;       //  BRGT
                 return TRUE;

        case 11: if (mode==0)
			destination = operand;                // d
                 else    destination = getoperand2(5,operand);// x
                 if (Status&VBIT)  PC = destination;         //  BRV
                 return TRUE;

        case 12: if (mode==0)
			destination = operand;                // d
                 else    destination = getoperand2(5,operand);// x
                 if (Status&CBIT)  PC = destination;          //  BRC
                 return TRUE;

        case 13: if (mode==0)
			destination = operand;                // d
                 else    destination = getoperand2(5,operand);// x
                 SP = SP-2;                                   //  CALL
                 if (SP<SPHWM) SPHWM=SP;
                 writetomem('d',SP,(PC&0xff00) >> 8 );
                 writetomem('d',SP+1,PC&0xFF );
                 PC = destination;
                 return TRUE;

        case 14: if (reg==0) tempa = RegA; else tempa = RegX;  // NOTr
                 tempa = ~tempa;
                 Status = 0;
                 if ( (tempa & 0xffff) == 0) Status |=ZBIT;
                 if ( (tempa & 0x8000) != 0) Status |=NBIT;
                 if (reg==0) RegA = tempa; else RegX = tempa;
                 return TRUE;

        case 15: if (reg==0) tempa = RegA; else tempa = RegX;  // NEGr
                 tempa = -tempa;
                 Status = 0;
                 if ( (tempa & 0xffff) == 0) Status |=ZBIT;
                 if ( (tempa & 0x8000) != 0) Status |=NBIT;
                 if ( (tempa & 0xffff) == 0x8000) Status |=VBIT;
                 if (reg==0) RegA = tempa; else RegX = tempa;
                 return TRUE;

       case 16:  if (reg==0) tempa = RegA; else tempa = RegX;  // ASLr
                 tempa = tempa << 1;
                 Status = 0;
                 if (tempa < -32768 || tempa > 32767) Status |= VBIT;
                 if ((tempa&0xFFFF) == 0) Status |= ZBIT;
                 if ((tempa&0x10000) != 0) Status |= CBIT;
                 if ((tempa&0x8000) != 0) Status |= NBIT;
                 if (reg==0) RegA = tempa&0xFFFF; else RegX = tempa&0xFFFF;
                 return TRUE;

       case 17:  if (reg==0) tempa = RegA; else tempa = RegX;  // ASRr
                 Status = 0;
                 if ((tempa&1) != 0) Status |= CBIT;
                 tempb = tempa >> 1;
                 if ((tempa&0x8000) != 0) 
                     {
                       tempb = tempb & 0x7FFF; tempb |= 0x8000;  // sign extension
                       Status |= NBIT;
                     }
                 if ((tempb&0xFFFF) == 0) Status |= ZBIT;
                 if (reg==0) RegA = tempb&0xFFFF; else RegX = tempb&0xFFFF;
                 return TRUE;

       case 18:  if (reg==0) tempa = RegA; else tempa = RegX;  // ROLr
                 Status = 0;
                 if ((tempa&0x8000) != 0) Status |= CBIT;
                 tempa = tempa << 1;
                 if (Status & CBIT) tempa += 1;
                 if (reg==0) RegA = tempa&0xFFFF; else RegX = tempa&0xFFFF;
                 return TRUE;

       case 19:  if (reg==0) tempa = RegA; else tempa = RegX;  // RORr
                 Status = 0;
                 if ((tempa&1) != 0) Status |= CBIT;
                 tempa = tempa >> 1;
                 tempa = tempa & 0x7fff;
                 if (Status & CBIT) tempa += 0x8000;
                 if (reg==0) RegA = tempa&0xFFFF; else RegX = tempa&0xFFFF;
                 return TRUE;
        
        case 20: return TRUE;                                        // NOP - unary

        case 21: destination = getoperand2(mode,operand);           
                 return TRUE;                                        // NOP - nonunary

	case 22: memdestination = effectiveaddress(mode,operand);      //DECI
                 printf("DECI: ");
                 fgets(buffer,50,stdin);
                 sscanf(buffer,"%d",&tempa);
                 Status = 0;
                 if (tempa<0) Status |= NBIT;
                 if (tempa==0) Status |= ZBIT;
                 if (tempa<-32768 || tempa > 32767) Status |=VBIT;
                 tempreg = tempa;                 
                 writetomem('d',memdestination,(tempreg&0xFF00)>>8);
                 writetomem('d',memdestination+1,tempreg&0xFF);
                 return TRUE;

        case 23: destination = getoperand2(mode,operand);
                 printf("%d",destination);                   // DECO - no trap version
                 return TRUE;

        case 24: destination = effectiveaddress(mode,operand);
                 tempchar = readfrommem('d',destination);    // STRO - no trap version
                 while (tempchar != 0)
                   {
                      printf("%c",tempchar);
                      destination++;
                      tempchar = readfrommem('d',destination);
                   }
                 return TRUE;

	case 25: memdestination = effectiveaddress(mode,operand);      //CHARI
                 printf("CHARI: ");
                 fgets(buffer,50,stdin);
                 sscanf(buffer,"%c",&inchar);
                 tempchar = inchar;                 
                 writetomem('d',memdestination,tempchar);
                 return TRUE;

        case 26: destination = getoperand1(mode,operand);
                 printf("%c",destination);          // CHARO
                 return TRUE;

        case 27: SP = SP + other;                           // RETN                           
                 tempa = readfrommem('d',SP);
                 SP++;
                 tempb = readfrommem('d',SP);
                 SP++;
                 tempa = tempa << 8;
                 PC = tempa + tempb;
                 return TRUE;

        case 28: destination = getoperand2(mode,operand);     // ADDSP
                 tempa = SP;
                 tempb = destination;
                 tempc = tempa+tempb;
                 Status = 0;
                 if (tempc&0x10000) Status |=CBIT;
                 if (tempb>0 && tempc<SP || tempb<0 && tempc>SP) Status |=VBIT;
                 SP = tempc&0xffff;
                 if (SP<SPHWM) SPHWM=SP;
                 if (SP == 0) Status |= ZBIT;
                 return TRUE;

        case 29: destination = getoperand2(mode,operand);     // SUBSP
                 tempa = SP;
                 tempb = destination;
                 Status = 0;
                 if ( (unsigned short)tempb > SP) Status |=CBIT;             
                 tempc = tempa-tempb;
                 if (tempb>0 && tempc>SP || tempb<0 && tempc<SP) Status |=VBIT;
                 SP = tempc&0xffff;
                 if (SP<SPHWM) SPHWM=SP;
                 if (SP == 0) Status |= ZBIT;
                 return TRUE;

        case 30: destination = getoperand2(mode,operand);     // ADD
                 if (reg==0) tempa = RegA; else tempa = RegX;
                 tempb = destination;
                 tempc = tempa+tempb;
                 Status = 0;
                 if (tempc&0x10000) Status |=CBIT;
                 if (tempc < -32768 || tempc > 32767) Status |=VBIT;
                 if (( tempc & 0xffff) ==0 ) Status |= ZBIT;
                 if (( tempc & 0x8000) !=0 ) Status |= NBIT;
                 if (reg==0) RegA=tempc; else RegX = tempc;
                 return TRUE;

        case 31: destination = getoperand2(mode,operand);     // SUBr
                 if (reg==0) tempa = RegA; else tempa = RegX;
                 tempb = destination;
                 Status = 0;
                 if ( (unsigned short)tempb > (unsigned short)tempa) Status |=CBIT;          
                 tempc = tempa-tempb;
                 if (tempc < -32768 || tempc > 32767) Status |=VBIT;
                 if (( tempc & 0xffff) ==0 ) Status |= ZBIT;
                 if (( tempc & 0x8000) !=0 ) Status |= NBIT;
                 if (reg==0) RegA=tempc; else RegX = tempc;
                 return TRUE;

        case 32: destination = getoperand2(mode,operand);     // ANDr
                 if (reg==0) tempa = RegA; else tempa = RegX;
                 tempb = destination;
                 tempc = tempa&tempb;
                 Status = 0;
                 if (( tempc & 0xffff) ==0 ) Status |= ZBIT;
                 if (( tempc & 0x8000) !=0 ) Status |= NBIT;
                 if (reg==0) RegA=tempc; else RegX = tempc;
                 return TRUE;

        case 33: destination = getoperand2(mode,operand);     // ORr
                 if (reg==0) tempa = RegA; else tempa = RegX;
                 tempb = destination;
                 tempc = tempa|tempb;
                 Status = 0;
                 if (( tempc & 0xffff) ==0 ) Status |= ZBIT;
                 if (( tempc & 0x8000) !=0 ) Status |= NBIT;
                 if (reg==0) RegA=tempc; else RegX = tempc;
                 return TRUE;

        case 34: destination = getoperand2(mode,operand);     // CPr
                 if (reg==0) tempa = RegA; else tempa = RegX;
                 tempb = destination;
                 Status = 0;
                 if ( (unsigned short)tempb > (unsigned short)tempa) Status |=CBIT;          
                 tempc = tempa-tempb;
                 if (tempc < -32768 || tempc > 32767) Status |=VBIT;
                 if (( tempc & 0xffff) ==0 ) Status |= ZBIT;
                 if (( tempc & 0x8000) !=0 ) Status |= NBIT;
                 return TRUE;

        case 35: destination = getoperand2(mode,operand);
                 if (reg==0) RegA = destination;             // LDr
                        else RegX = destination;
                 Status = 0;
                 if (destination<0) Status |= NBIT;
                 if (destination==0) Status |= ZBIT;
                 return TRUE;

        case 36: destination = getoperand1(mode,operand);  //LDBYTEr
                 Status = 0;
                 if (reg==0) 
                       {
 			 RegA = (RegA&0xFF00) | (destination&255); 
                         if (RegA<0)  Status |= NBIT;
                         if (RegA==0) Status |= ZBIT;
                       }
                    else 
                       {
                         RegX = (RegX&0xFF00) | (destination&255);
                         if (RegX<0)  Status |= NBIT;
                         if (RegX==0) Status |= ZBIT;
                       }
                 return TRUE;

	case 37: memdestination = effectiveaddress(mode,operand);      //STr
                 if (reg==0) tempreg = RegA; else tempreg = RegX;
                 writetomem('d',memdestination,(tempreg&0xFF00)>>8);
                 writetomem('d',memdestination+1,tempreg&0xFF);
                 return TRUE;

	case 38: memdestination = effectiveaddress(mode,operand);      // STBYTEr
                 if (reg==0) tempreg = RegA; else tempreg = RegX;
                 writetomem('d',memdestination,tempreg&0xFF);
                 return TRUE;

        default: printf("Not yet implemented: %d\n", decodetable[IR]);
                 return FALSE;
     }
}



void printregisters()
{
        if (Sysmode==USER) printf("\nMode = user "); else printf("\nMode = system");
	printf(" RegA:%.4hx RegX:%.4hx",RegA,RegX);
        printf(" SP:%.4hx PC:%.4hx Status:%.4hx",SP,PC,Status);
        printf(" NZVC: %1d%1d%1d%1d\n",
                ((Status&NBIT)==0) ? 0 : 1,
                ((Status&ZBIT)==0) ? 0 : 1,
                ((Status&VBIT)==0) ? 0 : 1,
                ((Status&CBIT)==0) ? 0 : 1);
}



void insertlist(int N)
{
  // add instance of N to the profile list

  PROFLISTNODE *temp, *two;

  if ( L == NULL || L->linenumber>N )
   {
     // new node at the beginning of the list
     if (( temp= (PROFLISTNODE*) malloc(sizeof(PROFLISTNODE))) == NULL)
            {
               printf("insertlist: cannot allocate memory for profile node\n");
               exit(1);
            }
     temp->linenumber=N;
     temp->count = 1;
     temp->next = L;
     L = temp;
     return;
   }

  if (L->linenumber == N) 
    {
      // increment count in first list node
      L->count = L->count+1;
      return;
    }

  two = L;
  while (two->next!=NULL && two->next->linenumber < N) two=two->next;
       if (two->next == NULL || two->next->linenumber > N)
            {  
              // insert or append new node
              temp= (PROFLISTNODE*) malloc(sizeof(PROFLISTNODE));
              temp->linenumber=N;
              temp->count = 1;
              temp->next = two->next;
              two->next=temp;
              return;
            }
       else two->next->count = two->next->count+1;  // add to count in existing node
}


void printlist()
{
 // prints profile list
 PROFLISTNODE *temp;
 temp=L;
 
 printf("Run-time profile\n");
 printf("Address\tAddress\tCount\n");
 printf(" (hex)\t (dec)\n");
 while (temp != NULL)
  {
   printf("%4x\t%4d\t%4d\n",temp->linenumber,temp->linenumber,temp->count);
   temp = temp->next;
  }

}

void deletelist()
{
  // delete profile list with storage recovery
  PROFLISTNODE *P;

  while (L !=NULL)
    {
       P = L;
       L = P-> next;
       free(P);
    }
}



void newrun()
{
 // modified version of run to permit correct operation of single-step

 int stopthisloop = FALSE ;
 int IR,opcode,istrap,unary,reg,mode,other;
 unsigned short temp; // temporary address
 short int operand;

 do {
      // fetch
      if ((tracenumber & TRACEBIT1) != 0) printf("PC = %x\n",PC);   
      insertlist(PC);   // profile list
      IR = readfrommem('i',PC); PC++;

      // decode
        if (!decode(IR,&opcode, &istrap, &unary, &reg, &mode, &other))
          {
           printf("bad decode %d\n",IR);
           stopthisloop = TRUE;
          }
        else {
               if ((tracenumber & TRACEBIT4) !=0) temptrace(opcode, istrap, unary,reg,mode,other,operand);
               if (!unary)
                {
                  operand = readfrommem('i',PC)<<8; PC++;
           	  operand = operand | readfrommem('i',PC); PC++;
                }

               //  execute
               if (istrap && ostrap)
                 {
                   // Trap action - see chapter 8
                   Sysmode = SYSTEM;
                   temp = readfrommem('d',0xfffa) << 8 | readfrommem('d',0xfffb); // pointer to system stack base
                   // printf("trapping: temp = %x\n",temp);
                   // printf("trapping: IR = %d, lsb(IR) = %d\n",IR,lsb(IR));
          	   writetomem('d',temp-1,lsb(IR));
                   writetomem('d',temp-3,msb(SP));    writetomem('d',temp-2,lsb(SP));
                   writetomem('d',temp-5,msb(PC));    writetomem('d',temp-4,lsb(PC));
                   writetomem('d',temp-7,msb(RegX));  writetomem('d',temp-6,lsb(RegX));
                   writetomem('d',temp-9,msb(RegA));  writetomem('d',temp-8,lsb(RegA));
                   writetomem('d',temp-10,Status);
                   SP = temp-10;
                   if (SP<SPHWM) SPHWM=SP;
                   PC = readfrommem('d',0xfffe) <<8 | readfrommem('d',0xffff);
                 }
               else
                 { 
                  // normal instruction
                  if (!execute(IR, istrap, unary,reg,mode,other,operand))
                   {
      	             printf("\nProgram Stop. PC= %d\n",PC);
                     if ((tracenumber & TRACEBIT2) != 0) printregisters();
                     stopthisloop=TRUE;
                   }
                  if ((tracenumber & TRACEBIT2) != 0) printregisters();
                  if (isinbreakpoints(PC)) printf("Breakpoint %x reached\n",PC);
                 }
              }
    } while (!singlestep && !stopthisloop && !isinbreakpoints(PC));
}




int run()
{
  int i,IR,opcode,istrap,unary,reg,mode,other, go_on = TRUE;
  short int operand;
  unsigned short temp;  // temporary address
  char throwaway;

  PC = 0;  RegA = 0; RegX = 0;  Status = 0;
  SP = (mem[0xFFF8]<<8)|mem[0xFFF9];
  SPHWM = SP;

  while (go_on)
   {
     // printf("*** top of loop in run PC = %d (%x) RegA = %d (%x) RegX = %d (%x) ***\n",PC,PC,RegA,RegA,RegX,RegX);
     if (singlestep)
         {
           printf(":");
           scanf("%c",&throwaway);
         }
     if (tracenumber>0) printf("Running:PC =  %d (%x)\n",PC,PC);
     IR = readfrommem('i',PC); PC++;
     if (!decode(IR,&opcode, &istrap, &unary, &reg, &mode, &other))
        {
           printf("bad decode %d\n",IR);
           return FALSE;
        }
     if (!unary)
           {
         	operand = readfrommem('i',PC)<<8; PC++;
           	operand = operand | readfrommem('i',PC); PC++;
           }
     if (istrap && ostrap)
        {
          // Trap action - see chapter 8
          Sysmode = SYSTEM;
          temp = readfrommem('d',0xfffa) << 8 | readfrommem('d',0xfffb);
         // printf("trapping: temp = %x\n",temp);
         // printf("trapping: IR = %d, lsb(IR) = %d\n",IR,lsb(IR));
          writetomem('d',temp-1,lsb(IR));
          writetomem('d',temp-3,msb(SP));    writetomem('d',temp-2,lsb(SP));
          writetomem('d',temp-5,msb(PC));    writetomem('d',temp-4,lsb(PC));
          writetomem('d',temp-7,msb(RegX));  writetomem('d',temp-6,lsb(RegX));
          writetomem('d',temp-9,msb(RegA));  writetomem('d',temp-8,lsb(RegA));
          writetomem('d',temp-10,Status);
          SP = temp-10;
          if (SP<SPHWM) SPHWM=SP;
          PC = readfrommem('d',0xfffe) <<8 | readfrommem('d',0xffff);
        }
     else
           { 
             // normal instruction
            if (tracenumber>1) temptrace(opcode, istrap, unary,reg,mode,other,operand);
            if (!execute(IR, istrap, unary,reg,mode,other,operand))
              {
      	        printf("\nProgram Stop. PC= %d\n",PC);
                if (tracenumber>2) printregisters();
                return TRUE;
              }
            if (tracenumber>2) printregisters();
          }
   }
}



int isinbreakpoints(unsigned int address)
{
 int i;

 for (i=0; i<numbreakpoints; i++)
   if (address==breakpoints[i]) return TRUE;

 return FALSE;
}



void breakpointhandler()
{
 // handles listing, adding and removing breakpoint addresses
 // submenu
 //              q quit
 //              l list
 //              a add
 //              r remove
 //              c clear

 char command;
 int i,j,toremove,found,runmore = TRUE;
 char buffer[50];

 while (runmore)
  {
   printf("(l)ist breakpoints;           (r)emove breakpoint n;\n");
   printf("(a)dd new breakpoint;         (c)lear breakpoint list;      (q)uit;\n");
   printf("breakpoint operation: ");
   fgets(buffer,50,stdin);
   sscanf(buffer,"%c",&command);
   switch (command)
     {
       case '?' :
       case 'h' : printf(" a: add  r: remove l: list c: clear q: quit\n");
                  break;
       case 'a' : if (numbreakpoints<MAXBREAKPOINTS)
                     {
                      printf("Breakpoint (hex): ");
                      fgets(buffer,50,stdin);
                      sscanf(buffer,"%x",&breakpoints[numbreakpoints]);
                      numbreakpoints++;
                     }
                  else
                      printf("cannot add: breakpoint list is full\n");
                  break;
       case 'c' : numbreakpoints = 0;
                  break;
       case 'l' : for (i=0; i<numbreakpoints; i++) printf("%4x  ",breakpoints[i]);
                  printf("\n");
                  break;
       case 'q' : runmore = FALSE;
                  break;
       case 'r' : printf("Which breakpoint to remove?: ");
                  fgets(buffer,50,stdin);
                  sscanf(buffer,"%x",&toremove);
                  found = FALSE;
                  for (i=0; i<numbreakpoints; i++)
                      {
                       if (breakpoints[i]==toremove)
                          {
                           found = TRUE;
                           numbreakpoints--;
                           for (j=i; j<numbreakpoints; j++) breakpoints[j]=breakpoints[j+1];
                          }
                      }
                  if (!found) printf("breakpoint %x not found\n",toremove);
                  break;
       default  : printf("breakpoint command not implemented %c\n",command);
     }
  }
}
   




void zeroout()
{
 int i;

 for (i=0; i<NUMINSTR; i++) instructioncounts[i]=0;
}


void printcachestats()
{
 printf(" Data reads=%5d\t Data read hits=%5d\tHR=%6.1f\n",dreadcount,dreadhits,(float)dreadhits/(float)dreadcount*100.0);
 printf("Inst. reads=%5d\tInst. read hits=%5d\tHR=%6.1f\n",ireadcount,ireadhits,(float)ireadhits/(float)ireadcount*100.0);
 printf("Total reads=%5d\tTotal read hits=%5d\tHR=%6.1f\n",
    (dreadcount+ireadcount),(dreadhits+ireadhits), 
    (float)(dreadhits+ireadhits)/(float)(dreadcount+ireadcount)*100.0);                                                                                                   
 printf("     Writes=%5d\t     Write hits=%5d\tHR=%6.1f\n",wcount,whits,(float)whits/(float)wcount*100.0);
}





void displayfreqs()
{
 int temp,total=0;

 char *names[]={"stop","rettr","movspa","movflga",
                "br","brle","brlt","breq","brne","brge",
                "brgt","brv","brc","call","notr","negr",
		 "aslr","asrr","rolr","rorr","nop(u)","nop",
                 "deci","deco","stro","chari","charo","retn",
                 "addsp","subsp","addr","subr","andr","orr",
                 "cpr","ldr","ldbyter","str","stbyter"
               };


 for (temp=0; temp<NUMINSTR; temp+=3)
       {
                   printf("%8s: %4d \t %8s: %4d \t %8s: %4d\n",
                          names[temp],instructioncounts[temp],
                          names[temp+1], instructioncounts[temp+1],
                          names[temp+2], instructioncounts[temp+2]
                         );
                   total += (instructioncounts[temp]+instructioncounts[temp+1]+instructioncounts[temp+2]);
       }
 printf("total = %d\n",total);
}



void menu()
{
 // display menu of options
    printf("a: reset PC and SP\n");
    printf("b: breakpoint options\n");
    printf("c: clear memory\n");
    printf("d: display memory\n");
    printf("e: extended display - registers\n");
    printf("f: instruction frequencies\n");
    printf("g: toggle single-step mode\n");
    printf("i: initialize cache\n");
    printf("l: load\n");
    printf("m: modify memory\n");
    printf("n: trace number (composite)\n");
    printf("   1 - PC   \n");
    printf("   2 - Register after instruction  \n");
    printf("   4 - opcode, mode,  register, other  \n");
    printf("   8 -  \n");
    printf("   16 - cache address mapping  \n");
    printf("   32 - cache changes   \n");
    printf("o: toggle OS trap of NOP,DECO,DECI,STRO\n");
    printf("p: toggle pctrace\n");
    printf("q: quit\n");
    printf("r: run\n");
    printf("s: cache stats\n");
    printf("t: toggle memtrace\n");
    printf("w: stack high water mark\n");
    printf("x: print profile list\n");
    printf("y: clear profile list\n");
    printf("z: zero out frequencies\n");
}


void shortmenu()
{
    printf("\n");
    printf("(l)oad;                      (r)un;\n");
    printf("(q)uit;                      (h)elp;\n");
    printf("(c)lear mem;                 (d)isplay mem;\n");
    printf("(m)odify mem;                (b)reakpoints;\n");
    printf("(i)nitialize cache;          cache (s)tats;\n");
    printf("trace (n)umber;              instruction (f)requencies;\n");
    printf("(z)ero out frequencies;      (e)xtended display;\n");
    printf("stack high (w)ater mark;     toggle sin(g)le-step mode;\n");
    printf("t(o)ggle OS trap;            toggle (p)c trace;\n");
    printf("toggle memory (t)race;       reset PC (a)nd SP;\n");
    printf("(x)print profile list;       (y)clear profile list;\n");
}


void initializecache()
{
 // clear tags, usage etc
 if (!iscache)
    {
      printf("initializecache: no cache created\n");
    }
 else
    {
      int i;
      for (i=0; i< totalblocks; i++) tags[i] = 0;  // arbitrary
      for (i=0; i< totalblocks; i++) inuse[i] = FALSE;
      for (i=0; i< totalblocks; i++) usage[i] = 0;  // good for all replacement algs?
      clocktime = 0;  // this clock only used in cache mechanism
      ireadcount = 0; ireadhits=0;
      dreadcount = 0; dreadhits=0;
      wcount=0; whits=0;
    }
}


void hardwaresetup()
{
 // prompts user for hardware configuration
 // if cache and what parameters

 char response;
 char algname[8];
 char buffer[50];


/*
 printf("Cache? (y/n): ");
 fgets(buffer,50,stdin);
 sscanf(buffer,"%c",&response);

 if (response=='n')
   {
       iscache = FALSE;    
   }
 else
*/
   {
       iscache = TRUE;
       // here set up cache global parameters and allocate global spaces
       printf("Split instruction and data caches? (y/n): ");  
       fgets(buffer,50,stdin);
       sscanf(buffer,"%c",&response);
       if (response=='n') sepid = FALSE; else sepid = TRUE;

       //printf("hardwaresetup: response=%c sepid=%d\n",response,sepid);

       printf("Block size is 2^N bytes. Enter N: "); 
       fgets(buffer,50,stdin); sscanf(buffer,"%d",&blockbits);

       printf("Enter number of blocks per row: "); 
       fgets(buffer,50,stdin); sscanf(buffer,"%d",&bpr);

       printf("Number of rows is 2^M. Enter M: "); 
       fgets(buffer,50,stdin); sscanf(buffer,"%d",&rowbits);

       do {
             printf("Replacement algorithm (random/fifo/lfu/lru): "); 
             fgets(buffer,50,stdin); sscanf(buffer,"%s",algname);
             algorithm = -1;
             if (strcmp(algname,"random")==0) algorithm=RANDOM;
             if (strcmp(algname,"fifo")==0) algorithm=FIFO;
             if (strcmp(algname,"lfu")==0) algorithm=LFU;
             if (strcmp(algname,"lru")==0) algorithm=LRU;
       
          }  while (algorithm == -1);

       blocksize = 1 << blockbits;
       numrows = 1 << rowbits;
       totalblocks = numrows * bpr;

       // allocate memory
                                                                      // added casts 9/8/05
       tags  = (unsigned int *) malloc(totalblocks * sizeof(int));
       inuse = (unsigned char*) malloc(totalblocks * sizeof(char));
       usage = (unsigned long*) malloc(totalblocks * sizeof(long));
       cache = (unsigned char*) malloc(totalblocks * blocksize * sizeof(char));
       // diagnostic 
       printf("blocksize=%d numrows=%d totalblocks=%d algorithm=%s\n", blocksize, numrows, totalblocks, algname);
  
   }

}

void warranty()
{
 //GPL warranty

printf("BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n");
printf("FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n");
printf("OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n");
printf("PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n");
printf("OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n");
printf("MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n");
printf("TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n");
printf("PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n");
printf("REPAIR OR CORRECTION.\n");
printf("\n");
printf("IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n");
printf("WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n");
printf("REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n");
printf("INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n");
printf("OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n");
printf("TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n");
printf("YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n");
printf("PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n");
printf("POSSIBILITY OF SUCH DAMAGES.\n");

}


void redistribute()
{
 // GPL redistribute conditions

printf("		    GNU GENERAL PUBLIC LICENSE\n");
printf("   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION\n");
printf("\n");
printf("  0. This License applies to any program or other work which contains\n");
printf("a notice placed by the copyright holder saying it may be distributed\n");
printf("under the terms of this General Public License.  The \"Program\", below,\n");
printf("refers to any such program or work, and a \"work based on the Program\"\n");
printf("means either the Program or any derivative work under copyright law:\n");
printf("that is to say, a work containing the Program or a portion of it,\n");
printf("either verbatim or with modifications and/or translated into another\n");
printf("language.  (Hereinafter, translation is included without limitation in\n");
printf("the term \"modification\".)  Each licensee is addressed as \"you\".\n");
printf("\n");
printf("Activities other than copying, distribution and modification are not\n");
printf("covered by this License; they are outside its scope.  The act of\n");
printf("running the Program is not restricted, and the output from the Program\n");
printf("is covered only if its contents constitute a work based on the\n");
printf("Program (independent of having been made by running the Program).\n");
printf("Whether that is true depends on what the Program does.\n");
printf("\n");
printf("  1. You may copy and distribute verbatim copies of the Program's\n");
printf("source code as you receive it, in any medium, provided that you\n");
printf("conspicuously and appropriately publish on each copy an appropriate\n");
printf("copyright notice and disclaimer of warranty; keep intact all the\n");
printf("notices that refer to this License and to the absence of any warranty;\n");
printf("and give any other recipients of the Program a copy of this License\n");
printf("along with the Program.\n");
printf("\n");
printf("You may charge a fee for the physical act of transferring a copy, and\n");
printf("you may at your option offer warranty protection in exchange for a fee.\n");
printf("\n");
printf("  2. You may modify your copy or copies of the Program or any portion\n");
printf("of it, thus forming a work based on the Program, and copy and\n");
printf("distribute such modifications or work under the terms of Section 1\n");
printf("above, provided that you also meet all of these conditions:\n");
printf("\n");
printf("    a) You must cause the modified files to carry prominent notices\n");
printf("    stating that you changed the files and the date of any change.\n");
printf("\n");
printf("    b) You must cause any work that you distribute or publish, that in\n");
printf("    whole or in part contains or is derived from the Program or any\n");
printf("    part thereof, to be licensed as a whole at no charge to all third\n");
printf("    parties under the terms of this License.\n");
printf("\n");
printf("    c) If the modified program normally reads commands interactively\n");
printf("    when run, you must cause it, when started running for such\n");
printf("    interactive use in the most ordinary way, to print or display an\n");
printf("    announcement including an appropriate copyright notice and a\n");
printf("    notice that there is no warranty (or else, saying that you provide\n");
printf("    a warranty) and that users may redistribute the program under\n");
printf("    these conditions, and telling the user how to view a copy of this\n");
printf("    License.  (Exception: if the Program itself is interactive but\n");
printf("    does not normally print such an announcement, your work based on\n");
printf("    the Program is not required to print an announcement.)\n");
printf("\n");
printf("These requirements apply to the modified work as a whole.  If\n");
printf("identifiable sections of that work are not derived from the Program,\n");
printf("and can be reasonably considered independent and separate works in\n");
printf("themselves, then this License, and its terms, do not apply to those\n");
printf("sections when you distribute them as separate works.  But when you\n");
printf("distribute the same sections as part of a whole which is a work based\n");
printf("on the Program, the distribution of the whole must be on the terms of\n");
printf("this License, whose permissions for other licensees extend to the\n");
printf("entire whole, and thus to each and every part regardless of who wrote it.\n");
printf("\n");
printf("Thus, it is not the intent of this section to claim rights or contest\n");
printf("your rights to work written entirely by you; rather, the intent is to\n");
printf("exercise the right to control the distribution of derivative or\n");
printf("collective works based on the Program.\n");
printf("\n");
printf("In addition, mere aggregation of another work not based on the Program\n");
printf("with the Program (or with a work based on the Program) on a volume of\n");
printf("a storage or distribution medium does not bring the other work under\n");
printf("the scope of this License.\n");
printf("\n");
printf("  3. You may copy and distribute the Program (or a work based on it,\n");
printf("under Section 2) in object code or executable form under the terms of\n");
printf("Sections 1 and 2 above provided that you also do one of the following:\n");
printf("\n");
printf("    a) Accompany it with the complete corresponding machine-readable\n");
printf("    source code, which must be distributed under the terms of Sections\n");
printf("    1 and 2 above on a medium customarily used for software interchange; or,\n");
printf("\n");
printf("    b) Accompany it with a written offer, valid for at least three\n");
printf("    years, to give any third party, for a charge no more than your\n");
printf("    cost of physically performing source distribution, a complete\n");
printf("    machine-readable copy of the corresponding source code, to be\n");
printf("    distributed under the terms of Sections 1 and 2 above on a medium\n");
printf("    customarily used for software interchange; or,\n");
printf("\n");
printf("    c) Accompany it with the information you received as to the offer\n");
printf("    to distribute corresponding source code.  (This alternative is\n");
printf("    allowed only for noncommercial distribution and only if you\n");
printf("    received the program in object code or executable form with such\n");
printf("    an offer, in accord with Subsection b above.)\n");
printf("\n");
printf("The source code for a work means the preferred form of the work for\n");
printf("making modifications to it.  For an executable work, complete source\n");
printf("code means all the source code for all modules it contains, plus any\n");
printf("associated interface definition files, plus the scripts used to\n");
printf("control compilation and installation of the executable.  However, as a\n");
printf("special exception, the source code distributed need not include\n");
printf("anything that is normally distributed (in either source or binary\n");
printf("form) with the major components (compiler, kernel, and so on) of the\n");
printf("operating system on which the executable runs, unless that component\n");
printf("itself accompanies the executable.\n");
printf("\n");
printf("If distribution of executable or object code is made by offering\n");
printf("access to copy from a designated place, then offering equivalent\n");
printf("access to copy the source code from the same place counts as\n");
printf("distribution of the source code, even though third parties are not\n");
printf("compelled to copy the source along with the object code.\n");
printf("\n");
printf("  4. You may not copy, modify, sublicense, or distribute the Program\n");
printf("except as expressly provided under this License.  Any attempt\n");
printf("otherwise to copy, modify, sublicense or distribute the Program is\n");
printf("void, and will automatically terminate your rights under this License.\n");
printf("However, parties who have received copies, or rights, from you under\n");
printf("this License will not have their licenses terminated so long as such\n");
printf("parties remain in full compliance.\n");
printf("\n");
printf("  5. You are not required to accept this License, since you have not\n");
printf("signed it.  However, nothing else grants you permission to modify or\n");
printf("distribute the Program or its derivative works.  These actions are\n");
printf("prohibited by law if you do not accept this License.  Therefore, by\n");
printf("modifying or distributing the Program (or any work based on the\n");
printf("Program), you indicate your acceptance of this License to do so, and\n");
printf("all its terms and conditions for copying, distributing or modifying\n");
printf("the Program or works based on it.\n");
printf("\n");
printf("  6. Each time you redistribute the Program (or any work based on the\n");
printf("Program), the recipient automatically receives a license from the\n");
printf("original licensor to copy, distribute or modify the Program subject to\n");
printf("these terms and conditions.  You may not impose any further\n");
printf("restrictions on the recipients' exercise of the rights granted herein.\n");
printf("You are not responsible for enforcing compliance by third parties to\n");
printf("this License.\n");
printf("\n");
printf("  7. If, as a consequence of a court judgment or allegation of patent\n");
printf("infringement or for any other reason (not limited to patent issues),\n");
printf("conditions are imposed on you (whether by court order, agreement or\n");
printf("otherwise) that contradict the conditions of this License, they do not\n");
printf("excuse you from the conditions of this License.  If you cannot\n");
printf("distribute so as to satisfy simultaneously your obligations under this\n");
printf("License and any other pertinent obligations, then as a consequence you\n");
printf("may not distribute the Program at all.  For example, if a patent\n");
printf("license would not permit royalty-free redistribution of the Program by\n");
printf("all those who receive copies directly or indirectly through you, then\n");
printf("the only way you could satisfy both it and this License would be to\n");
printf("refrain entirely from distribution of the Program.\n");
printf("\n");
printf("If any portion of this section is held invalid or unenforceable under\n");
printf("any particular circumstance, the balance of the section is intended to\n");
printf("apply and the section as a whole is intended to apply in other\n");
printf("circumstances.\n");
printf("\n");
printf("It is not the purpose of this section to induce you to infringe any\n");
printf("patents or other property right claims or to contest validity of any\n");
printf("such claims; this section has the sole purpose of protecting the\n");
printf("integrity of the free software distribution system, which is\n");
printf("implemented by public license practices.  Many people have made\n");
printf("generous contributions to the wide range of software distributed\n");
printf("through that system in reliance on consistent application of that\n");
printf("system; it is up to the author/donor to decide if he or she is willing\n");
printf("to distribute software through any other system and a licensee cannot\n");
printf("impose that choice.\n");
printf("\n");
printf("This section is intended to make thoroughly clear what is believed to\n");
printf("be a consequence of the rest of this License.\n");
printf("\n");
printf("  8. If the distribution and/or use of the Program is restricted in\n");
printf("certain countries either by patents or by copyrighted interfaces, the\n");
printf("original copyright holder who places the Program under this License\n");
printf("may add an explicit geographical distribution limitation excluding\n");
printf("those countries, so that distribution is permitted only in or among\n");
printf("countries not thus excluded.  In such case, this License incorporates\n");
printf("the limitation as if written in the body of this License.\n");
printf("\n");
printf("  9. The Free Software Foundation may publish revised and/or new versions\n");
printf("of the General Public License from time to time.  Such new versions will\n");
printf("be similar in spirit to the present version, but may differ in detail to\n");
printf("address new problems or concerns.\n");
printf("\n");
printf("Each version is given a distinguishing version number.  If the Program\n");
printf("specifies a version number of this License which applies to it and \"any\n");
printf("later version\", you have the option of following the terms and conditions\n");
printf("either of that version or of any later version published by the Free\n");
printf("Software Foundation.  If the Program does not specify a version number of\n");
printf("this License, you may choose any version ever published by the Free Software\n");
printf("Foundation.\n");
printf("\n");
printf("  10. If you wish to incorporate parts of the Program into other free\n");
printf("programs whose distribution conditions are different, write to the author\n");
printf("to ask for permission.  For software which is copyrighted by the Free\n");
printf("Software Foundation, write to the Free Software Foundation; we sometimes\n");
printf("make exceptions for this.  Our decision will be guided by the two goals\n");
printf("of preserving the free status of all derivatives of our free software and\n");
printf("of promoting the sharing and reuse of software generally.\n");
printf("\n");
}


void outputgnu()
{
 // output GNU GLP stuff

 printf("Pep8Cache version %s, Copyright (C) 2006 Peter D. Smith\n",VERSION);
 printf("Pep8Cache comes with ABSOLUTELY NO WARRANTY; for details type u at the Pep8Cache prompt.\n");
 printf("This is free software, and you are welcome to redistribute it under certain conditions;\n");
 printf("for details type v at the Pep8Cache prompt.\n\n");
}



int main()
{
 int temp,runmore=TRUE,start,finish,menuonprompt;
 char command,first,second;
 int addresstomod;
 unsigned char newbyte;
 char buffer[50];
 char loadfilename[50];
 int memtracefileopen = FALSE;
 memtrace = FALSE; instructiontrace=FALSE, pctrace = FALSE; tracenumber = 0; ostrap=FALSE; singlestep=FALSE;
 Sysmode = USER;
 iscache = FALSE;
 clearmem();
 for (temp=0; temp<NUMINSTR; temp++) instructioncounts[temp]=0;
 numbreakpoints = 0;
 L = NULL;  // profilelist

 outputgnu();

 temp= loados();
 if (temp) printf("OS loaded OK\n"); 
      else {
             printf("unable to load OS. Aborting\n");
             exit(1);
           }

 printf("Menu on prompt (y/n)? ");
 fgets(buffer,50,stdin);
 sscanf(buffer,"%c",&command);
 if (command=='y') menuonprompt=TRUE; else menuonprompt=FALSE;

 hardwaresetup();         // set up optional cache
 initializecache();

 while (runmore)
  {
    if (menuonprompt) shortmenu();
    printf("Pep8Cache: ");

   /* bits in trace level:
   1 - PC   
   2 - Register after instruction  
   4 - opcode, mode,  register, other  
   8 -  
   16 - cache address mapping  
   32 - cache changes   
*/
    fgets(buffer,50,stdin);
    sscanf(buffer,"%c",&command);
    switch(command)
     {
      case 'h':
      case '?': menu();
                break;
      case 'q': runmore = FALSE;
                break;
      case 'a': PC = 0;
                SP = (mem[0xFFF8]<<8)|mem[0xFFF9];  // OS pointer to base of user stack.
                SPHWM = SP;
                break;
      case 'b': breakpointhandler();
                break;
      case 'c': clearmem();
                break;
      case 'f': displayfreqs();
                break;
      case 'i': initializecache();
                break;
      case 'l': printf("load file name: ");
                fgets(buffer,50,stdin);
                sscanf(buffer,"%s",loadfilename);
                temp = load(loadfilename);
                if (temp==0) printf("bad load\n"); else printf("%d  bytes loaded\n",temp);
                break;
      case 'm': printf("input address to modify (hex): ");
                fgets(buffer,50,stdin);
                sscanf(buffer,"%x",&addresstomod);
                if (addresstomod < WRITELIMIT)
                	{
                		printf("input new contents (xx): ");
                                fgets(buffer,50,stdin);
                		sscanf(buffer,"%c%c",&first,&second);
                		temp = convert(first,second);
                		if (temp != -1) mem[addresstomod]=temp;
				else
				printf("error: bad contents %c%c\n",first,second);
			}
		else printf("error: address outside Application Program area\n");
                break;
      case 'n': printf("00 0001 - PC\n");
                printf("00 0010 - Register after instruction\n");
                printf("00 0100 - opcode, mode, register, other\n");
                printf("00 1000 - \n");
                printf("01 0000 - cache address mapping\n");
                printf("10 0000 - cache changes\n");
                printf("trace number (hex) (%x): ",tracenumber);
                fgets(buffer,50,stdin);
		sscanf(buffer,"%x",&tracenumber);
                break;
      case 'o': if (ostrap) ostrap=FALSE; else ostrap=TRUE;
                if (ostrap) printf("ostrap on\n"); else printf("ostrap off\n");
                break; 
      case 'g': if (singlestep) singlestep=FALSE; else singlestep=TRUE;
                if (singlestep) printf("single-step on\n"); else printf("single-step off\n");
                break; 
      case 'p': if (pctrace) pctrace=FALSE; else pctrace=TRUE;
                if (pctrace) printf("pctrace on\n"); else printf("pctrace off\n");
                break; 
      case 'd': printf("enter bounds of display (hex): ");
                fgets(buffer,50,stdin);
                sscanf (buffer,"%x%x",&start,&finish);
                display(start,finish);
                break;
      case 'e': printregisters();
                break;
      case 't': if (memtrace) memtrace=FALSE; else memtrace=TRUE;
                if (memtrace) printf("memtrace on\n"); else printf("memtrace off\n");
                if (memtrace && !memtracefileopen)
                   {
                     memtracefile = fopen("memtrace","w");
                     if (memtracefile != NULL) memtracefileopen = TRUE;
                   }
                break;
      case 's': printcachestats(); 
                break;
      case 'w': printf("SP high water mark = %4x\n",SPHWM);
                break;
      case 'r': newrun();
                break;
      case 'u': warranty();
                break;
      case 'v': redistribute();
                break;
      case 'x': printlist();  
                break;
      case 'y': deletelist();
                printf("profile list deleted\n");
                break;
      case 'z': zeroout();
                break;
      default : printf("command %c not implemented\n",command);
     }
  }

 if (memtrace) fclose(memtracefile);
 if (instructiontrace) fclose(instructiontracefile);

 printf("Pep8Cache end\n");
}


